--[[----------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--



	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'

	--WF.config
local LrmConfig = require 'LrmConfig'

local WFLogger = require "WFLogger"

--------------------------------------------------------------------------------

local logger = WFLogger "LriSyncStatusModel"
--	:setLevel 'debug'


local derefOpt = KSTableUtils.derefOpt

--============================================================================--

local LrmSyncStatusModel = {}
local Model = {}

--------------------------------------------------------------------------------

local function getUploadProgress( catalog )

	local numNeedingUpload = catalog:indexStore():query {

		sql = [[SELECT
			count( DISTINCT docId )
		FROM
			missingBinariesOnOz
		WHERE
			binaryType = "master"
		]],
		calm = 2,
		quiesceUntilUpToDate = true
	}

	:map( function( progressRows )

		local binaryUpload = derefOpt( progressRows, 1, 1 ) or 0

		return binaryUpload

	end )

	return numNeedingUpload

end

--------------------------------------------------------------------------------

local function getDetailedUploadProgress( catalog )

	-- SPAM [immisch 2019-07-19]: We can't just look at sumBytesNeedToUpload and sumFileSize
	-- because when binaries disappear from missingBinariesOnOz, the progress gets restarted
	-- which leads to the % upladed actually going down.

	local sumFileSizes = 0

	local numNeedingUpload = catalog:indexStore():query {

		sql = [[
			SELECT
				T1.sumBytesNeedToUpload, T2.sumFileSize
			FROM
				(SELECT SUM( bytesNeedToUpload ) AS sumBytesNeedToUpload 
				FROM missingBinariesOnOz) T1
				, 
				(SELECT SUM( fileSize ) AS sumFileSize 
				FROM ( SELECT 
							DISTINCT docId, fileSize 
						FROM 
							missingBinariesOnOz
				)) T2
		]],
	}
		
	:map( function( progressRows )
	
		local bytesNeedToUpload = derefOpt( progressRows, 1, 1 )
		local fileSizes = derefOpt( progressRows, 1, 2 )
		
		if not bytesNeedToUpload or not fileSizes then
		
			sumFileSizes = 0
		
			return nil
		else
		
			if fileSizes > sumFileSizes then
				sumFileSizes = fileSizes
			end
		
		end
		
		local sumBytesUploaded = sumFileSizes - bytesNeedToUpload
		
		logger:debugf( "Upload complete: %d/%d %f", sumBytesUploaded, sumFileSizes, sumBytesUploaded / sumFileSizes * 100.0 )

		return {
			bytesUploaded = sumBytesUploaded,
			totalFileSize = sumFileSizes
		}
					
	end )
	
	return numNeedingUpload
	
end

--------------------------------------------------------------------------------

-- @mark Sub models

--------------------------------------------------------------------------------

function Model:makeBinaryDownloadProgressModel( binaryDownloadStrategy )

	binaryDownloadStrategy = binaryDownloadStrategy or LrmConfig.StrategyPreferProxy

	local sql = [[SELECT
					count( DISTINCT r.docId )
				FROM
					renditionRevisions as r
				INNER JOIN
					album_asset
				ON
					album_asset.assetId = r.docId
				WHERE
					album_asset.albumId IN
					(
						SELECT
							docId
						FROM
							albums
						WHERE
							useOffline = 1
					)
				
					AND
					
					renditionName IN %s
				]]

	if binaryDownloadStrategy == LrmConfig.StrategyPreferMaster then
		sql = string.format( sql, "( 'thumbnail2x', 'preview', 'proxyWithoutMaster', 'master' )" )
	else
		sql = string.format( sql, "( 'thumbnail2x', 'preview', 'proxy' )" )
	end

	local progress = self._catalog:indexStore():query( { sql = sql } )
		:map( function( rows )

			return rows[ 1 ][ 1 ]

		end )


	return {
		value = progress,
	}

end

--------------------------------------------------------------------------------

function Model:makeBinaryDownloadProgressAndAlbumModel( binaryDownloadStrategy )

	binaryDownloadStrategy = binaryDownloadStrategy or LrmConfig.StrategyPreferProxy

	local sql = [[SELECT albums.name, count( DISTINCT r.docId ) FROM
					renditionRevisions as r
				INNER JOIN
					album_asset
				ON
					album_asset.assetId = r.docId
				INNER JOIN 
					albums
				ON 
					albums.docId = album_asset.albumId and albums.useOffline = 1
				WHERE
					renditionName IN %s
				]]

	if binaryDownloadStrategy == LrmConfig.StrategyPreferMaster then
		sql = string.format( sql, "( 'thumbnail2x', 'preview', 'proxyWithoutMaster', 'master', 'xmp_develop' )" )
	else
		sql = string.format( sql, "( 'thumbnail2x', 'preview', 'proxy', 'xmp_develop' )" )
	end

	local progress = self._catalog:indexStore():query( { sql = sql } )
		:map( function( rows )

			return { 
				album = rows[ 1 ][ 1 ], 
				count = rows[ 1 ][ 2 ] }

		end )


	return {
		value = progress,
	}

end

--------------------------------------------------------------------------------

function LrmSyncStatusModel.makeModel( catalog )

	assert( catalog )

	local uploadingToOzProgress = getUploadProgress( catalog )

	return {
		_catalog = catalog,

		value = {
			isUploading = catalog:getSyncClient():isPushingMetadata(),
			isDownloading = catalog:getSyncClient():isPullingMetadata(),
			pendingUploadsToOz = uploadingToOzProgress,
			progressBinaryUploads = getDetailedUploadProgress( catalog ),
		},

		models = Model,

		events = {
		},

		close = function( self )


		end
	}

end

--------------------------------------------------------------------------------

return LrmSyncStatusModel
