--[[----------------------------------------------------------------------------

WFCatSyncAutoDeleter.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'
local Observable = Rx.Observable

	-- WF.core.utils
local KSFunctionUtils = require 'KSFunctionUtils'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFRxRunUntilDone = require 'WFRxRunUntilDone'
local WFRxScheduler = require 'WFRxScheduler'

	-- WF.oz_client.managed_catalog
local WFCatDelete = require 'WFCatDelete'

	-- WF.core.logger
local WFLogger = require 'WFLogger'

--------------------------------------------------------------------------------

local logger = WFLogger( 'WFCatSyncAutoDeleter' )

local check = assert( WFArguments.check )

local runUntilDone = WFRxRunUntilDone

local KSTableUtils_derefOpt = assert( KSTableUtils.derefOpt )

--------------------------------------------------------------------------------

local kAutoDeleterChangeCounterLocalDocId = 'WFCatSyncAutoDeleter-changeCounter'

--============================================================================--

-- @mark WFCatSyncAutoDeleter class

--------------------------------------------------------------------------------

local WFCatSyncAutoDeleter = {

	_docStore = false,
		-- (WFRxDocStore)
	
	_indexStore = false,
		-- (WFRxIndexStore)

	_docIdsToProcess = false,
		-- (table)

	_deletedDocIdStreamSubscription = false,
		-- (Rx Subscription)
		
	_processDeletionsStreamSubscription = false,
		-- (Rx Subscription)

	_processQueriesSubscription = false,
		-- (Rx Subscription)
		
	_continueFrom = false,
		-- (number)
		
	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------

local getDocIdsAffectedByDeletedAlbumSQL = KSFunctionUtils.memoizeFn( function( numberOfValues )
	
	return string.format( "SELECT docId FROM album_asset_v2 WHERE albumId IN ( ?%s )",
			string.rep( ", ?", numberOfValues - 1 ) )

end )

--------------------------------------------------------------------------------

local getDocIdsAffectedByDeletedAssetSQL = KSFunctionUtils.memoizeFn( function( numberOfValues )
	
	return string.format( "SELECT docId FROM album_asset_v2 WHERE assetId IN ( ?%s )",
			string.rep( ", ?", numberOfValues - 1 ) )

end )

--------------------------------------------------------------------------------

local function makeChunkedDeletionStream( makeGetDeletionIdsSQLFunc, docStore, indexStore, docIdsToDelete, startOffset, chunkSize )

	local idsToDeleteSQL = makeGetDeletionIdsSQLFunc( chunkSize )

	logger:infof( "Building deletion query: %s", idsToDeleteSQL )

	return indexStore:queryOnce {
			sql = idsToDeleteSQL,		 
			quiesceUntilUpToDate = true,		 
			unpack( docIdsToDelete, startOffset, startOffset + chunkSize - 1 )
		}
		
		:flatMapLatest( function( docIdRows )
		
			return WFCatDelete.deleteDocsFromQueryResult( docStore, docIdRows, true )		
			
		end )		  			

end

--------------------------------------------------------------------------------

local function deleteRelatedDocuments( ... )

	return makeChunkedDeletionStream( getDocIdsAffectedByDeletedAlbumSQL, ... ),
		makeChunkedDeletionStream( getDocIdsAffectedByDeletedAssetSQL, ... )

end

--------------------------------------------------------------------------------

local function processDeletedDocIds( self )

	local docIds = self._docIdsToProcess
	local numberOfDocsToProcess = #docIds
	
	local continueFrom = self._continueFrom
		-- It's possible for self._continueFrom to be modified after this function
		-- is called, but before the observables within this function complete. 
		-- We need to make sure use the continueFrom value that was present when
		-- this fn was called, hence the local.
		
	local offset = 1
	
	local queries = {}
	
	while offset <= numberOfDocsToProcess do
	
		local chunkSize = 0
	
		if numberOfDocsToProcess >= 32 then
		
			chunkSize = 32
		
		elseif numberOfDocsToProcess >= 16 then
		
			chunkSize = 16
		
		elseif numberOfDocsToProcess >= 8 then
		
			chunkSize = 8
		
		elseif numberOfDocsToProcess >= 4 then
		
			chunkSize = 4
		
		elseif numberOfDocsToProcess >= 2 then
		
			chunkSize = 2
		
		else
		
			chunkSize = 1
		
		end
	
		logger:infof( "Now deleting %d docs", chunkSize )
	
		local docIdsAffectedByDeletedAlbumQuery, docIdsAffectedByDeletedAssetQuery 
				= deleteRelatedDocuments( self._docStore, self._indexStore, docIds, offset, chunkSize )
	
		queries[ #queries + 1 ] = docIdsAffectedByDeletedAlbumQuery
		
		queries[ #queries + 1 ] = docIdsAffectedByDeletedAssetQuery
	
		offset = offset + chunkSize
	
	end
	
	if #queries > 0 then
		
		local disposeFn = function()
		
			self._processQueriesSubscription:dispose()
			self._processQueriesSubscription = nil
		
		end

		self._processQueriesSubscription = Observable.from( queries )
		
			:map( function( query )

				return query
				
			end )
			
			:flatMerge( 1 )
				-- this prevents the queries from running concurrently.
		
			:subscribe( nil, 
				function( maybeErr )
				
					logger:errorf( "An error occured %s", maybeErr ) 
					disposeFn()
					
				end, 
				function()
				
					self._docStore:updateDeletedDocsSinceNamedChangeCounter( 
							kAutoDeleterChangeCounterLocalDocId, continueFrom )
						:call( runUntilDone )
					
					disposeFn()
					
				end )

	end

	self._docIdsToProcess = {}
	
end

--------------------------------------------------------------------------------

function WFCatSyncAutoDeleter:__init( catalog, optArgs )

	check( 'WFCatSyncAutoDeleter', 'WFCatalog', '?table' )

	optArgs = optArgs or {}

	self._docStore = catalog:docStore()
	self._docStore:addCloseTarget( self )

	self._indexStore = catalog:indexStore()
	self._indexStore:addCloseTarget( self )

	self._docIdsToProcess = {}
	
	-- Queue up the deleted document IDs to process

	local docStore = self._docStore
	
	self._deletedDocIdStreamSubscription = docStore
	
		:deletedDocsSinceNamedChangeCounter( kAutoDeleterChangeCounterLocalDocId )

		:subscribe( function( docValue, continueFrom )
				
			-- Prevent the feeding of the id to WFCatSyncAutoDeleter if we're
			-- looking at an asset that has been moved to the trash, identifiable
			-- by 'new_id'. In the case of albums, we want the id to get passed
			-- through, because albums don't go to the trash, they are immediately
			-- deleted, and then we also need the album assets cleaned up. Which
			-- is what WFCatSyncAutoDeleter does.
			
			local new_id = KSTableUtils_derefOpt( docValue, '_serverProvidedMetadata', 'new_id' )
			
			if not ( docValue.type == 'asset' and new_id ) then
			
				table.insert( self._docIdsToProcess, docValue._id )
				
			end
			
			self._continueFrom = continueFrom
			
		end )

	-- Every 15 seconds, process the deleted document IDs.

	self._processDeletionsStreamSubscription = Observable.timer( 0, optArgs.delay or 15, WFRxScheduler.sharedScheduler() )
		
		:filter( function()
			return not self._processQueriesSubscription
				-- If we're already processing deleted doc ids, wait.
		end )

		:subscribe( function()
			processDeletedDocIds( self )
		end )
		
end

--------------------------------------------------------------------------------

function WFCatSyncAutoDeleter:close()

	check( 'WFCatSyncAutoDeleter' )
	
	local deletedDocIdStreamSubscription = self._deletedDocIdStreamSubscription
	self._deletedDocIdStreamSubscription = nil
	
	if deletedDocIdStreamSubscription then
		deletedDocIdStreamSubscription:dispose()
	end

	local processDeletionsStreamSubscription = self._processDeletionsStreamSubscription
	self._processDeletionsStreamSubscription = nil
	
	if processDeletionsStreamSubscription then
		processDeletionsStreamSubscription:dispose()
	end
	
	local processQueriesSubscription = self._processQueriesSubscription
	self._processQueriesSubscription = nil
	
	if processQueriesSubscription then
		processQueriesSubscription:dispose()
	end

end

--------------------------------------------------------------------------------

return Rx.Class 'WFCatSyncAutoDeleter'( WFCatSyncAutoDeleter )
